@@ -3,6 +3,7 @@ gdk = gtk.gdk
 from math import pi, sin, cos, sqrt, atan2
 import struct
 import cairo
+import windowing
 from lib.helpers import rgb_to_hsv, hsv_to_rgb, clamp
 import colorhistory as ch
 from gettext import gettext as _
@@ -11,7 +12,7 @@ CSTEP=0.007
 PADDING=4
 
 class GColorSelector(gtk.DrawingArea):
-    def __init__(self, init_dnd=True):
+    def __init__(self):
         gtk.DrawingArea.__init__(self)
         self.color = (1,0,0)
         self.hsv = (0,1,1)
@@ -28,18 +29,25 @@ class GColorSelector(gtk.DrawingArea):
         self.connect('button-release-event', self.on_button_release)
         self.connect('configure-event', self.on_configure)
         self.connect('motion-notify-event', self.motion)
-        if init_dnd:
-            self.connect('drag_data_get', self.drag_get)
         self.button_pressed = False
-        self.do_select = True
         self.grabbed = False
-        self.dnd_enabled = init_dnd
-        self.set_size_request(80, 80)
+        self.set_size_request(110, 100)
 
-    def test_drag(self,x,y,d):
-        return d>20
+    def test_move(self, x, y):
+        """
+        Return true during motion if the selection indicator
+        should move to where the cursor is pointing.
+        """
+        return True
 
-    def move(self,x,y):
+    def test_button_release(self, x, y):
+        """
+        Return true after a button-release-event if the colour under the
+        pointer should be selected.
+        """
+        return True
+
+    def move(self, x, y):
         pass
 
     def redraw_on_select(self):
@@ -48,31 +56,12 @@ class GColorSelector(gtk.DrawingArea):
     def motion(self,w, event):
         if not self.button_pressed:
             return
-        d = sqrt((event.x-self.press_x)**2 + (event.y-self.press_y)**2)
-        if self.dnd_enabled and self.test_drag(event.x,event.y, d):
-            self.do_select = False
-            self.drag_begin([("application/x-color",0,80)], gdk.ACTION_COPY, 1, event)
-        else:
+        if self.test_move(event.x, event.y):
             if not self.grabbed:
                 self.grabbed = True
                 self.grab_add()
             self.move(event.x,event.y)
 
-    def drag_get(self, widget, context, selection, targetType, eventTime):
-        try:
-            color, is_hsv = self.get_color_at(self.press_x, self.press_y)
-        except TypeError:
-            return
-        if is_hsv:
-            r,g,b = hsv_to_rgb(*color)
-        else:
-            r,g,b = color
-        r = min(int(r*65536), 65535)
-        g = min(int(g*65536), 65535)
-        b = min(int(b*65536), 65535)
-        clrs = struct.pack('HHHH',r,g,b,0)
-        selection.set(selection.target, 8, clrs)
-
     def on_configure(self,w, size):
         self.x_rel = size.x
         self.y_rel = size.y
@@ -116,13 +105,12 @@ class GColorSelector(gtk.DrawingArea):
 
     def on_button_press(self,w, event):
         self.button_pressed = True
-        self.do_select = True
         self.press_x = event.x
         self.press_y = event.y
         self.record_button_press(event.x, event.y)
 
     def on_button_release(self,w, event):
-        if self.button_pressed and self.do_select:
+        if self.button_pressed and self.test_button_release(event.x, event.y):
             self.select_color_at(event.x,event.y)
         self.button_pressed = False
         if self.grabbed:
@@ -172,10 +160,14 @@ class RecentColors(gtk.HBox):
 
 CIRCLE_N = 12.0
 SLOTS_N = 5
+NEUTRAL_MID_GREY  = (0.5, 0.5, 0.5)  #: Background neutral mid grey, intended to permit colour comparisons without theme colours distracting and imposing themselves
+NEUTRAL_DARK_GREY = (0.46, 0.46, 0.46) #: Slightly distinct outlinefor colour pots, intended to reduce "1+1=3 (or more)" effects
 
-AREA_SQUARE = 1
-AREA_SAMPLE = 2
-AREA_CIRCLE = 3
+AREA_SQUARE = 1  #: Central saturation/value square
+AREA_INSIDE = 2  #: Grey area inside the rings
+AREA_SAMPLE = 3  #: Inner ring of color sampler boxes (hue)
+AREA_CIRCLE = 4  #: Outer gradiated ring (hue)
+AREA_OUTSIDE = 5  #: Blank outer area, outer space
 
 sq32 = sqrt(3)/2
 sq33 = sqrt(3)/3
@@ -203,6 +195,10 @@ class CircleSelector(GColorSelector):
         self.analogous = False
         self.square = False
 
+    def has_harmonies_visible(self):
+        return self.complimentary or self.triadic or self.double_comp \
+          or self.split_comp or self.analogous or self.square
+
     def get_previous_color(self):
         return self.color
 
@@ -214,15 +210,20 @@ class CircleSelector(GColorSelector):
         return x1,y1,x2,y2
 
     def configure_calc(self):
+        padding = 5
         self.x0 = self.x_rel + self.w/2.0
         self.y0 = self.y_rel + self.h/2.0
-        self.M = M = min(self.w,self.h)-2
-        self.rd = 0.31*M       # gives size of square
-        self.r2 = 0.42*M      # Inner radius of hue ring
-        self.r3 = M/2.0
+        self.M = M = min(self.w,self.h) - (2*padding)
+        self.r3 = 0.5*M     # outer radius of hue ring
+        if not self.has_harmonies_visible():
+            self.r2 = 0.42*M  # inner radius of hue ring
+            self.rd = 0.42*M  # and size of square
+        else:
+            self.r2 = 0.44*M  # line between hue ring & harmony samplers
+            self.rd = 0.34*M  # size of square & inner edge of harmony samplers
         self.m = self.rd/sqrt(2.0)
         self.circle_img = None
-        self.stroke_width = 0.015*M
+        self.stroke_width = 0.01*M
 
     def set_color(self, color, redraw=True):
         self.color = color
@@ -232,11 +233,29 @@ class CircleSelector(GColorSelector):
         if redraw:
             self.queue_draw()
 
-    def test_drag(self,x,y,dist):
-        area = self.area_at(self.press_x,self.press_y)
-        return area == AREA_SAMPLE
+    def test_move(self, x, y):
+        from_area = self.area_at(self.press_x,self.press_y)
+        to_area = self.area_at(x, y)
+        if from_area == AREA_CIRCLE:
+            return to_area not in (AREA_SQUARE, AREA_INSIDE)
+        elif from_area == AREA_SQUARE:
+            return to_area not in (AREA_CIRCLE, AREA_OUTSIDE)
+        else:
+            return False
+        #return from_area == to_area
+        #    and from_area in [AREA_CIRCLE, AREA_SQUARE]
 
-    def nearest_dragable(self, x,y):
+    def test_button_release(self, x, y):
+        from_area = self.area_at(self.press_x,self.press_y)
+        to_area = self.area_at(x, y)
+        return from_area == to_area \
+            and from_area in [AREA_SAMPLE, AREA_CIRCLE, AREA_SQUARE]
+
+    def nearest_move_target(self, x,y):
+        """
+        Returns the nearest (x, y) the selection indicator can move to during a
+        move with the button held down.
+        """
         area_source = self.area_at(self.press_x, self.press_y)
         area = self.area_at(x,y)
         if area == area_source and area in [AREA_CIRCLE, AREA_SQUARE]:
@@ -258,19 +277,23 @@ class CircleSelector(GColorSelector):
         return rx,ry
 
     def move(self,x,y):
-        x_,y_ = self.nearest_dragable(x,y)
+        x_,y_ = self.nearest_move_target(x,y)
         self.select_color_at(x_,y_)
 
     def area_at(self, x,y):
         dx = x-self.x0
         dy = y-self.y0
         d = sqrt(dx*dx+dy*dy)
-        if self.r2 < d < self.r3:
+        if d > self.r3:
+            return AREA_OUTSIDE
+        elif self.r2 < d < self.r3:
             return AREA_CIRCLE
         elif self.rd < d < self.r2:
             return AREA_SAMPLE
         elif abs(dx)<self.m and abs(dy)<self.m:
             return AREA_SQUARE
+        else:
+            return AREA_INSIDE
 
     def get_color_at(self, x,y):
         area = self.area_at(x,y)
@@ -295,15 +318,21 @@ class CircleSelector(GColorSelector):
     def draw_square(self,width,height,radius):
         img = cairo.ImageSurface(cairo.FORMAT_ARGB32, width,height)
         cr = cairo.Context(img)
-        h,s,v = self.hsv
         m = radius/sqrt(2.0)
+
+        # Slightly darker border for the colour area
+        sw = max(0.75*self.stroke_width, 3.0)
+        cr.set_source_rgb(*NEUTRAL_DARK_GREY)
+        cr.rectangle(self.x0-m-sw, self.y0-m-sw, 2*(m+sw), 2*(m+sw))
+        cr.fill()
+
+        h,s,v = self.hsv
         ds = 2*m*CSTEP
         v = 0.0
         x1 = self.x0-m
         x2 = self.x0+m
         y = self.y0-m
         while v < 1.0:
-            y += ds
             g = cairo.LinearGradient(x1,y,x2,y)
             g.add_color_stop_rgb(0.0, *hsv_to_rgb(h,0.0,1.0-v))
             g.add_color_stop_rgb(1.0, *hsv_to_rgb(h,1.0,1.0-v))
@@ -311,6 +340,7 @@ class CircleSelector(GColorSelector):
             cr.rectangle(x1,y, 2*m, ds)
             cr.fill_preserve()
             cr.stroke()
+            y += ds
             v += CSTEP
         h,s,v = self.hsv
         x = self.x0-m + s*2*m
@@ -318,6 +348,7 @@ class CircleSelector(GColorSelector):
         cr.set_source_rgb(*hsv_to_rgb(1-h,1-s,1-v))
         cr.arc(x,y, 3.0, 0.0, 2*pi)
         cr.stroke()
+
         return img
 
     def small_circle(self, cr, size, color, angle, rad):
@@ -399,17 +430,27 @@ class CircleSelector(GColorSelector):
         r,g,b = rgb
         return 1-r,1-g,1-b
 
-    def draw_inside_circle(self,width,height):
+    def draw_central_fill(self, cr, r):
+        """Draws a neutral-coloured grey circle of the specified radius in the central area."""
+        cr.set_source_rgb(*NEUTRAL_MID_GREY)
+        cr.arc(self.x0, self.y0, r, 0, 2*pi)
+        cr.fill()
+
+    def draw_harmony_ring(self,width,height):
+        """Draws the harmony ring if any colour harmonies are visible."""
         if not self.window:
             return
         points_size = min(width,height)/27.
         img = cairo.ImageSurface(cairo.FORMAT_ARGB32, width,height)
         cr = cairo.Context(img)
+        if not self.has_harmonies_visible():
+            self.draw_central_fill(cr, self.r2)
+            return img
         h,s,v = self.hsv
         a = h*2*pi
         self.angles = []
         self.simple_colors = []
-        cr.set_line_width(self.stroke_width)
+        cr.set_line_width(0.75*self.stroke_width)
         self.samples = [self.hsv]
         for i in range(int(CIRCLE_N)):
             c1 = c = h + i/CIRCLE_N
@@ -429,7 +470,7 @@ class CircleSelector(GColorSelector):
             cr.arc(self.x0, self.y0, self.r2, a1, a2)
             cr.line_to(self.x0,self.y0)
             cr.fill_preserve()
-            cr.set_source_rgb(0.5,0.5,0.5)
+            cr.set_source_rgb(*NEUTRAL_DARK_GREY) # "lines" between samples
             cr.stroke()
             # Indicate harmonic colors
             if self.triadic and i%(CIRCLE_N/3)==0:
@@ -451,24 +492,33 @@ class CircleSelector(GColorSelector):
             if self.analogous and i in [0,1,CIRCLE_N-1]:
                 self.small_rect(cr, points_size, self.inv(clr), an, (self.r2+self.rd)/2)
                 try_put(self.samples, hsv)
+        # Fill the centre
+        self.draw_central_fill(cr, self.rd)
+        # And an inner thin line
+        cr.set_source_rgb(*NEUTRAL_DARK_GREY)
+        cr.arc(self.x0, self.y0, self.rd, 0, 2*pi)
+        cr.stroke()
+        return img
+
+    def draw_circle_indicator(self, cr):
+        """Draws the indicator which shows the current hue on the outer ring."""
+        h, s, v = self.hsv
+        a = h*2*pi
         x1 = self.x0 + self.r2*cos(-a)
         y1 = self.y0 + self.r2*sin(-a)
         x2 = self.x0 + self.r3*cos(-a)
         y2 = self.y0 + self.r3*sin(-a)
         self.last_line = x1,y1, x2,y2, h
         cr.set_line_width(0.8*self.stroke_width)
-        cr.set_source_rgb(0,0,0)
+        cr.set_source_rgb(0, 0, 0)
         cr.move_to(x1,y1)
         cr.line_to(x2,y2)
         cr.stroke()
-        cr.set_source_rgb(0.5,0.5,0.5)
-        cr.arc(self.x0, self.y0, self.rd, 0, 2*pi)
-        cr.fill()
-        return img
 
     def draw_circles(self, cr):
-        cr.set_line_width(self.stroke_width)
-        cr.set_source_rgb(0.5,0.5,0.5)
+        """Draws two grey lines just inside and outside the outer hue ring."""
+        cr.set_line_width(0.75*self.stroke_width)
+        cr.set_source_rgb(*NEUTRAL_DARK_GREY)
         cr.arc(self.x0,self.y0, self.r2, 0, 2*pi)
         cr.stroke()
         cr.arc(self.x0,self.y0, self.r3, 0, 2*pi)
@@ -479,7 +529,7 @@ class CircleSelector(GColorSelector):
             return self.circle_img
         img = cairo.ImageSurface(cairo.FORMAT_ARGB32, w,h)
         cr = cairo.Context(img)
-        cr.set_line_width(0.8*self.stroke_width)
+        cr.set_line_width(0.75*self.stroke_width)
         a1 = 0.0
         while a1 < 2*pi:
             clr = hsv_to_rgb(a1/(2*pi), 1.0, 1.0)
@@ -492,17 +542,7 @@ class CircleSelector(GColorSelector):
         self.circle_img = img
         return img
 
-    def draw(self,w,event):
-        if not self.window:
-            return
-        cr = self.window.cairo_create()
-        cr.set_source_surface(self.draw_circle(self.w,self.h))
-        cr.paint()
-        cr.set_source_surface(self.draw_inside_circle(self.w,self.h))
-        cr.paint()
-        self.draw_circles(cr)
-        cr.set_source_surface(self.draw_square(self.w,self.h, self.rd*0.92))
-        cr.paint()
+    def draw_prev_vs_current_semicircles(self, cr):
         sq2 = sqrt(2.0)
         M = self.M/2.0
         r = (sq2-1)*M/(2*sq2)
@@ -515,13 +555,27 @@ class CircleSelector(GColorSelector):
         cr.arc(x0, y0, 0.9*r, pi/2, 3*pi/2)
         cr.fill()
         cr.arc(x0, y0, 0.9*r, 0, 2*pi)
-        cr.set_source_rgb(0.5,0.5,0.5)
-        cr.set_line_width(0.8*self.stroke_width)
+        cr.set_source_rgb(*NEUTRAL_DARK_GREY)
+        cr.set_line_width(0.75*self.stroke_width)
         cr.stroke()
 
+    def draw(self,w,event):
+        if not self.window:
+            return
+        cr = self.window.cairo_create()
+        cr.set_source_surface(self.draw_circle(self.w,self.h))
+        cr.paint()
+        cr.set_source_surface(self.draw_harmony_ring(self.w,self.h))
+        cr.paint()
+        self.draw_circle_indicator(cr)
+        self.draw_circles(cr)
+        cr.set_source_surface(self.draw_square(self.w, self.h, self.rd*0.95))
+        cr.paint()
+        #self.draw_prev_vs_current_semicircles(cr)
+
 class VSelector(GColorSelector):
     def __init__(self, color=(1,0,0), height=16):
-        GColorSelector.__init__(self, init_dnd=False)
+        GColorSelector.__init__(self)
         self.color = color
         self.hsv = rgb_to_hsv(*color)
         self.set_size_request(height*2,height)
@@ -653,6 +707,7 @@ def make_spin(min,max, changed_cb):
     adj = gtk.Adjustment(0,min,max, 1,10)
     btn = gtk.SpinButton(adj)
     btn.connect('value-changed', changed_cb)
+    btn.set_sensitive(False)
     return btn
 
 class HSVSelector(gtk.VBox):
@@ -796,7 +851,10 @@ class RGBSelector(gtk.VBox):
         self.set_color((r,g, spin.get_value()/255.))
 
 class Selector(gtk.VBox):
-    def __init__(self,app):
+
+    CIRCLE_MIN_SIZE = (200, 200)
+
+    def __init__(self, app, window):
         gtk.VBox.__init__(self)
         self.app = app
         hbox = gtk.HBox()
@@ -805,8 +863,14 @@ class Selector(gtk.VBox):
         hbox.pack_start(vbox,expand=True)
         self.recent = RecentColors()
         self.circle = CircleSelector()
+        self.circle.set_size_request(*self.CIRCLE_MIN_SIZE)
         vbox.pack_start(self.circle, expand=True)
 
+        self.window_ = window
+        self.window_.connect("configure-event", self.window_configured_cb)
+        self.window_.connect("map-event", self.window_mapped_cb)
+        self.window_mapped = False
+
         self.rgb_selector = RGBSelector()
         self.hsv_selector = HSVSelector()
         self.rgb_selector.on_select = self.rgb_selected
@@ -815,15 +879,21 @@ class Selector(gtk.VBox):
         nb.append_page(self.rgb_selector, gtk.Label(_('RGB')))
         nb.append_page(self.hsv_selector, gtk.Label(_('HSV')))
 
+        # Colour history
         self.exp_history = expander = gtk.Expander(_('Colors history'))
         expander.set_spacing(6)
         expander.add(self.recent)
+        expander.connect("notify::expanded", self.expander_expanded_cb, 'history')
         self.pack_start(expander, expand=False)
+
+        # Colour details
         self.exp_details = expander = gtk.Expander(_('Details'))
         expander.set_spacing(6)
         expander.add(nb)
+        expander.connect("notify::expanded", self.expander_expanded_cb, 'details')
         self.pack_start(expander, expand=False)
 
+        # Colour scheme harmonies
         def harmony_checkbox(attr, label):
             cb = gtk.CheckButton(label)
             cb.connect('toggled', self.harmony_toggled, attr)
@@ -854,7 +924,9 @@ class Selector(gtk.VBox):
         vbox_exp = gtk.VBox()
         vbox_exp.pack_start(frame1)
         expander.add(vbox_exp)
+        expander.connect("notify::expanded", self.expander_expanded_cb, 'harmonies')
         self.pack_start(expander, expand=False)
+
         self.circle.on_select = self.hue_selected
         self.circle.get_previous_color = self.previous_color
         self.recent.on_select = self.recent_selected
@@ -875,6 +947,7 @@ class Selector(gtk.VBox):
 
     def harmony_toggled(self, checkbox, attr):
         setattr(self.circle, attr, not getattr(self.circle, attr))
+        self.circle.configure_calc()
         self.queue_draw()
 
     def previous_color(self):
@@ -907,28 +980,58 @@ class Selector(gtk.VBox):
     def on_select(self,color):
         pass
 
-class Window(gtk.Window):
+    def expander_expanded_cb(self, expander, junk, cfg_stem):
+        if not self.window_mapped:
+            return
+        # After expander activation (both expand and un-expand), don't let the
+        # colour circle shrink smaller than its current size, and force the WM
+        # to shrink-fit the window to the new size.
+        circle_alloc = self.circle.get_allocation()
+        self.circle.set_size_request(circle_alloc.width, circle_alloc.height)
+        self.window_.set_resizable(False)
+        self.app.preferences['colorsampler.%s_expanded' % cfg_stem] \
+          = bool(expander.get_expanded())
+
+    def window_configured_cb(self, window, event, *param):
+        if not self.window_mapped:
+            return
+        # Undo the expander trick if it's still in effect: the user should be
+        # able to resize the dialog.
+        if not self.window_.get_resizable():
+            self.window_.set_resizable(True)
+            self.circle.set_size_request(*self.CIRCLE_MIN_SIZE)
+
+    def window_mapped_cb(self, window, event, *params):
+        # On the first map event, restore expander state from prefs.
+        if self.window_mapped:
+            return
+        if self.app.preferences.get("colorsampler.history_expanded", False):
+            self.exp_history.set_expanded(True)
+        if self.app.preferences.get("colorsampler.details_expanded", False):
+            self.exp_details.set_expanded(True)
+        if self.app.preferences.get("colorsampler.harmonies_expanded", False):
+            self.exp_config.set_expanded(True)
+        self.window_mapped = True
+
+class Window(windowing.SubWindow):
     def __init__(self,app):
-        gtk.Window.__init__(self)
-        self.app = app
+        windowing.SubWindow.__init__(self, app)
         self.set_title(_('MyPaint color selector'))
         self.set_role('Color selector')
-        self.set_default_size(270,300)
         self.connect('delete-event', self.app.hide_window_cb)
-        self.selector = Selector(app)
+        self.selector = Selector(app, self)
         self.selector.on_select = self.on_select
         self.exp_history = self.selector.exp_history
         self.exp_details = self.selector.exp_details
         self.exp_config = self.selector.exp_config
-        # TODO: persistency
-        #self.exp_history.set_expanded(str_to_bool( app.get_config('State', 'color_history_expanded') ))
-        #self.exp_details.set_expanded(str_to_bool( app.get_config('State', 'color_details_expanded') ))
-        #self.exp_config.set_expanded(str_to_bool( app.get_config('State', 'color_configure_expanded') ))
+
         self.add(self.selector)
-        self.app.kbm.add_window(self)
         self.app.brush.settings_observers.append(self.brush_modified_cb)
         self.stop_callback = False
 
+        # The first callback notification happens before the window is initialized
+        self.selector.set_color(app.brush.get_color_rgb())
+
     def brush_modified_cb(self):
         self.stop_callback = True
         self.selector.set_color(self.app.brush.get_color_rgb())
@@ -937,6 +1040,4 @@ class Window(gtk.Window):
     def on_select(self, color):
         if self.stop_callback:
             return
-        self.app.colorSelectionWindow.set_color_rgb(color)
         self.app.brush.set_color_rgb(color)
-
