@@ -17,6 +17,15 @@ from gettext import ngettext
 from lib import document, helpers
 import drawwindow
 
+import mimetypes
+
+SAVE_FORMAT_ANY = 0
+SAVE_FORMAT_ORA = 1
+SAVE_FORMAT_PNGSOLID = 2
+SAVE_FORMAT_PNGTRANS = 3
+SAVE_FORMAT_PNGMULTI = 4
+SAVE_FORMAT_JPEG = 5
+
 # Utility function to work around the fact that gtk FileChooser/FileFilter
 # does not have an easy way to use case insensitive filters
 def get_case_insensitive_glob(string):
@@ -25,6 +34,20 @@ def get_case_insensitive_glob(string):
     globlist = ["[%s%s]" % (c.lower(), c.upper()) for c in ext]
     return '*.%s' % ''.join(globlist)
 
+def add_filters_to_dialog(filters, dialog):
+    for name, patterns in filters:
+        f = gtk.FileFilter()
+        f.set_name(name)
+        for p in patterns:
+            f.add_pattern(get_case_insensitive_glob(p))
+        dialog.add_filter(f)
+
+def dialog_set_filename(dialog, s):
+    # According to pygtk docu we should use set_filename(),
+    # however doing so removes the selected filefilter.
+    path, name = os.path.split(s)
+    dialog.set_current_folder(path)
+    dialog.set_current_name(name)
 
 class FileHandler(object):
     def __init__(self, app):
@@ -39,9 +62,10 @@ class FileHandler(object):
         ('Reload',       gtk.STOCK_REFRESH, _('Reload'), 'F5', None, self.reload_cb),
         ('Save',         gtk.STOCK_SAVE, _('Save'), '<control>S', None, self.save_cb),
         ('SaveAs',       gtk.STOCK_SAVE_AS, _('Save As...'), '<control><shift>S', None, self.save_as_cb),
+        ('Export',       gtk.STOCK_SAVE_AS, _('Export...'), '<control><shift>E', None, self.save_as_cb),
         ('SaveScrap',    None, _('Save As Scrap'), 'F2', None, self.save_scrap_cb),
-        ('PrevScrap',    None, _('Open Previous Scrap'), 'F5', None, self.open_scrap_cb),
-        ('NextScrap',    None, _('Open Next Scrap'), 'F6', None, self.open_scrap_cb),
+        ('PrevScrap',    None, _('Open Previous Scrap'), 'F6', None, self.open_scrap_cb),
+        ('NextScrap',    None, _('Open Next Scrap'), 'F7', None, self.open_scrap_cb),
         ]
         ag = gtk.ActionGroup('FileActions')
         ag.add_actions(file_actions)
@@ -64,11 +88,41 @@ class FileHandler(object):
         self.active_scrap_filename = None
         self.set_recent_items()
 
+        self.file_filters = [ #(name, patterns)
+        (_("All Recognized Formats"), ("*.ora", "*.png", "*.jpg", "*.jpeg")),
+        (_("OpenRaster (*.ora)"), ("*.ora",)),
+        (_("PNG (*.png)"), ("*.png",)),
+        (_("JPEG (*.jpg; *.jpeg)"), ("*.jpg", "*.jpeg")),
+        ]
+        self.saveformats = [ #(name, extension, options)
+        (_("By extension (prefer default format)"), None, {}), #0
+        (_("OpenRaster (*.ora)"), '.ora', {}), #1
+        (_("PNG solid with background (*.png)"), '.png', {'alpha': False}), #2
+        (_("PNG transparent (*.png)"), '.png', {'alpha': True}), #3
+        (_("Multiple PNG transparent (*.XXX.png)"), '.png', {'multifile': True}), #4
+        (_("JPEG 90% quality (*.jpg; *.jpeg)"), '.jpg', {'quality': 90}), #5
+        ]
+        self.ext2saveformat = {
+        '.ora': SAVE_FORMAT_ORA, 
+        '.png': SAVE_FORMAT_PNGSOLID, 
+        '.jpeg': SAVE_FORMAT_JPEG, 
+        '.jpg': SAVE_FORMAT_JPEG}
+        self.config2saveformat = {
+        'openraster': SAVE_FORMAT_ORA,
+        'jpeg-90%': SAVE_FORMAT_JPEG,
+        'png-solid': SAVE_FORMAT_PNGSOLID,
+        }
+
     def set_recent_items(self):
         # this list is consumed in open_last_cb
+
+        # Note: i.exists() does not work on Windows if the pathname
+        # contains utf-8 characters. Since GIMP also saves its URIs
+        # with utf-8 characters into this list, I assume this is a
+        # gtk bug.  So we use our own test instead of i.exists().
         self.recent_items = [
                 i for i in gtk.recent_manager_get_default().get_items()
-                if "mypaint" in i.get_applications() and i.exists()
+                if "mypaint" in i.get_applications() and os.path.exists(helpers.uri2filename(i.get_uri()))
         ]
         self.recent_items.reverse()
 
@@ -86,38 +140,45 @@ class FileHandler(object):
     filename = property(get_filename, set_filename)
 
     def init_save_dialog(self):
-        dialog = gtk.FileChooserDialog(_("Save.."), self.app.drawWindow,
+        dialog = gtk.FileChooserDialog(_("Save..."), self.app.drawWindow,
                                        gtk.FILE_CHOOSER_ACTION_SAVE,
                                        (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                         gtk.STOCK_SAVE, gtk.RESPONSE_OK))
         self.save_dialog = dialog
         dialog.set_default_response(gtk.RESPONSE_OK)
         dialog.set_do_overwrite_confirmation(True)
+        add_filters_to_dialog(self.file_filters, dialog)
 
-        filter2info = {}
-        self.filter2info = filter2info
+        # Add widget for selecting save format
+        box = gtk.HBox()
+        label = gtk.Label(_('Format to save as:'))
+        label.set_alignment(0.0, 0.0)
+        combo = self.saveformat_combo = gtk.combo_box_new_text()
+        for name, ext, opt in self.saveformats:
+            combo.append_text(name)
+        combo.set_active(0)
+        combo.connect('changed', self.selected_save_format_changed_cb)
+        box.pack_start(label)
+        box.pack_start(combo, expand=False)
+        dialog.set_extra_widget(box)
+        dialog.show_all()
 
-        filters = [ #name, patterns, saveopts
-        (_("Any format (prefer OpenRaster)"), ("*.ora", "*.png", "*.jpg", "*.jpeg"), ('.ora', {})),
-        (_("OpenRaster (*.ora)"), ("*.ora", ), ('.ora', {})),
-        ("PNG solid with background (*.png)", ("*.png", ), ('.png', {'alpha': False})),
-        (_("PNG transparent (*.png)"), ("*.png", ), ('.png', {'alpha': True})),
-        (_("Multiple PNG transparent (*.XXX.png)"), ("*.png", ), ('.png', {'multifile': True})),
-        (_("JPEG 90% quality (*.jpg; *.jpeg)"), ("*.jpg", "*.jpeg"), ('.jpg', {'quality': 90})),
-        ]
-        for nr, filt in enumerate(filters):
-            name, patterns, saveopts = filt
-            f = gtk.FileFilter()
-            filter2info[f] = saveopts
-            f.set_name(name)
-            for pat in patterns:
-                f.add_pattern(get_case_insensitive_glob(pat))
-            dialog.add_filter(f)
-            if nr == 0:
-                self.save_filter_default = f
+    def selected_save_format_changed_cb(self, widget):
+        """When the user changes the selected format to save as in the dialog, 
+        change the extension of the filename (if existing) immediately."""
+        dialog = self.save_dialog
+        filename = dialog.get_filename()
+        if filename:
+            filename = filename.decode('utf-8')
+            filename, ext = os.path.splitext(filename)
+            if ext:
+                saveformat = self.saveformat_combo.get_active()
+                ext = self.saveformats[saveformat][1]
+                if ext is not None:
+                    dialog_set_filename(dialog, filename+ext)
 
-    def confirm_destructive_action(self, title='Confirm', question='Really continue?'):
-        t = self.doc.unsaved_painting_time
+    def confirm_destructive_action(self, title=_('Confirm'), question=_('Really continue?')):
+        t = self.doc.model.unsaved_painting_time
         if t < 30:
             # no need to ask
             return True
@@ -153,44 +214,61 @@ class FileHandler(object):
     def new_cb(self, action):
         if not self.confirm_destructive_action():
             return
-        bg = self.doc.background
-        self.doc.clear()
-        self.doc.set_background(bg)
+        bg = self.doc.model.background
+        self.doc.model.clear()
+        self.doc.model.set_background(bg)
         self.filename = None
         self.set_recent_items()
-        self.app.drawWindow.reset_view_cb(None)
+        self.app.doc.reset_view_cb(None)
 
     @drawwindow.with_wait_cursor
     def open_file(self, filename):
         try:
-            self.doc.load(filename)
+            self.doc.model.load(filename)
         except document.SaveLoadError, e:
             self.app.message_dialog(str(e),type=gtk.MESSAGE_ERROR)
         else:
             self.filename = os.path.abspath(filename)
             print 'Loaded from', self.filename
-            self.app.drawWindow.reset_view_cb(None)
+            self.app.doc.reset_view_cb(None)
 
     @drawwindow.with_wait_cursor
-    def save_file(self, filename, **options):
+    def save_file(self, filename, export=False, **options):
         try:
-            x, y, w, h =  self.doc.get_bbox()
+            x, y, w, h =  self.doc.model.get_bbox()
             if w == 0 and h == 0:
                 raise document.SaveLoadError, _('Did not save, the canvas is empty.')
-            self.doc.save(filename, **options)
+            self.doc.model.save(filename, **options)
         except document.SaveLoadError, e:
             self.app.message_dialog(str(e),type=gtk.MESSAGE_ERROR)
         else:
-            self.filename = os.path.abspath(filename)
-            print 'Saved to', self.filename
-            gtk.recent_manager_get_default().add_full("file://" + self.filename,
-                    {
-                        'app_name': 'mypaint',
-                        'app_exec': sys.argv[0],
-                        # todo: get mime_type
-                        'mime_type': 'application/octet-stream'
-                    }
-            )
+            if not export:
+                self.filename = os.path.abspath(filename)
+                print 'Saved to', self.filename
+                gtk.recent_manager_get_default().add_full(helpers.filename2uri(self.filename),
+                        {
+                            'app_name': 'mypaint',
+                            'app_exec': sys.argv[0],
+                            # todo: get mime_type
+                            'mime_type': 'application/octet-stream'
+                        }
+                )
+            else:
+                print 'Exported to', os.path.abspath(filename)
+
+    def update_preview_cb(self, file_chooser, preview):
+        filename = file_chooser.get_preview_filename()
+        if filename:
+            filename = filename.decode('utf-8')
+            pixbuf = helpers.get_freedesktop_thumbnail(filename)
+            if pixbuf:
+                # if pixbuf is smaller than 128px in width, copy it onto a transparent 128x128 pixbuf
+                pixbuf = helpers.pixbuf_thumbnail(pixbuf, 128, 128, True)
+                preview.set_from_pixbuf(pixbuf)
+                file_chooser.set_preview_widget_active(True)
+            else:
+                #TODO display "no preview available" image
+                pass
 
     def open_cb(self, action):
         if not self.confirm_destructive_action():
@@ -201,25 +279,28 @@ class FileHandler(object):
                                         gtk.STOCK_OPEN, gtk.RESPONSE_OK))
         dialog.set_default_response(gtk.RESPONSE_OK)
 
-        filters = [ #name, patterns
-        (_("All Recognized Formats"), ("*.ora", "*.png", "*.jpg", "*.jpeg")),
-        (_("OpenRaster (*.ora)"), ("*.ora",)),
-        (_("PNG (*.png)"), ("*.png",)),
-        (_("JPEG (*.jpg; *.jpeg)"), ("*.jpg", "*.jpeg")),
-        ]
-        for name, patterns in filters:
-            f = gtk.FileFilter()
-            f.set_name(name)
-            for p in patterns:
-                f.add_pattern(get_case_insensitive_glob(p))
-            dialog.add_filter(f)
+        preview = gtk.Image()
+        dialog.set_preview_widget(preview)
+        dialog.connect("update-preview", self.update_preview_cb, preview)
+
+        add_filters_to_dialog(self.file_filters, dialog)
 
         if self.filename:
             dialog.set_filename(self.filename)
+        else:
+            # choose the most recent save folder
+            self.set_recent_items()
+            for item in reversed(self.recent_items):
+                uri = item.get_uri()
+                fn = helpers.uri2filename(uri)
+                dn = os.path.dirname(fn)
+                if os.path.isdir(dn):
+                    dialog.set_current_folder(dn)
+                    break
         try:
             if dialog.run() == gtk.RESPONSE_OK:
                 dialog.hide()
-                self.open_file(dialog.get_filename())
+                self.open_file(dialog.get_filename().decode('utf-8'))
         finally:
             dialog.destroy()
 
@@ -233,38 +314,44 @@ class FileHandler(object):
         if not self.save_dialog:
             self.init_save_dialog()
         dialog = self.save_dialog
-
-        def dialog_set_filename(s):
-            # According to pygtk docu we should use set_filename(),
-            # however doing so removes the selected filefilter.
-            path, name = os.path.split(s)
-            dialog.set_current_folder(path)
-            dialog.set_current_name(name)
-
         if self.filename:
-            dialog_set_filename(self.filename)
+            dialog_set_filename(dialog, self.filename)
         else:
-            dialog_set_filename('')
-            dialog.set_filter(self.save_filter_default)
+            dialog_set_filename(dialog, '')
             # choose the most recent save folder
             self.set_recent_items()
             for item in reversed(self.recent_items):
                 uri = item.get_uri()
-                fn = helpers.get_file_path_from_dnd_dropped_uri(uri)
+                fn = helpers.uri2filename(uri)
                 dn = os.path.dirname(fn)
                 if os.path.isdir(dn):
                     dialog.set_current_folder(dn)
                     break
 
         try:
+            # Loop until we have filename with an extension
             while dialog.run() == gtk.RESPONSE_OK:
-
-                filename = dialog.get_filename()
+                filename = dialog.get_filename().decode('utf-8')
                 name, ext = os.path.splitext(filename)
-                ext_filter, options = self.filter2info.get(dialog.get_filter(), ('ora', {}))
+                saveformat = self.saveformat_combo.get_active()
 
+                # If no explicitly selected format, use the extension to figure it out
+                if saveformat == SAVE_FORMAT_ANY:
+                    cfg = self.app.preferences['saving.default_format']
+                    default_saveformat = self.config2saveformat[cfg]
+                    if ext:
+                        try: 
+                            saveformat = self.ext2saveformat[ext]
+                        except KeyError:
+                            saveformat = default_saveformat
+                    else:
+                            saveformat = default_saveformat
+
+                desc, ext_format, options = self.saveformats[saveformat]
+
+                # 
                 if ext:
-                    if ext_filter != ext:
+                    if ext_format != ext:
                         # Minor ugliness: if the user types '.png' but
                         # leaves the default .ora filter selected, we
                         # use the default options instead of those
@@ -272,14 +359,17 @@ class FileHandler(object):
                         options = {}
                     assert(filename)
                     dialog.hide()
-                    self.save_file(filename, **options)
+                    if action.get_name() == 'Export':
+                        # Do not change working file
+                        self.save_file(filename, True, **options)
+                    else:
+                        self.save_file(filename, **options)
                     break
 
-                # add proper extension
-                filename = name + ext_filter
+                filename = name + ext_format
 
                 # trigger overwrite confirmation for the modified filename
-                dialog_set_filename(filename)
+                dialog_set_filename(dialog, filename)
                 dialog.response(gtk.RESPONSE_OK)
 
         finally:
@@ -289,6 +379,11 @@ class FileHandler(object):
         filename = self.filename
         prefix = self.get_scrap_prefix()
 
+        # If necessary, create the folder(s) the scraps are stored under
+        prefix_dir = os.path.dirname(prefix)
+        if not os.path.exists(prefix_dir): 
+            os.makedirs(prefix_dir)
+
         number = None
         if filename:
             l = re.findall(re.escape(prefix) + '([0-9]+)', filename)
@@ -319,17 +414,17 @@ class FileHandler(object):
                     maximum = number
             filename = '%s%03d_a' % (prefix, maximum+1)
 
-        #if self.doc.is_layered():
-        #    filename += '.ora'
-        #else:
-        #    filename += '.png'
-        filename += '.ora'
+        # Add extension
+        cfg = self.app.preferences['saving.default_format']
+        default_saveformat = self.config2saveformat[cfg]
+        filename += self.saveformats[default_saveformat][1]
 
         assert not os.path.exists(filename)
         self.save_file(filename)
 
     def get_scrap_prefix(self):
-        prefix = self.app.settingsWindow.save_scrap_prefix
+        prefix = self.app.preferences['saving.scrap_prefix']
+        prefix = os.path.expanduser(prefix.decode('utf-8'))
         prefix = os.path.abspath(prefix)
         if os.path.isdir(prefix):
             if not prefix.endswith(os.path.sep):
@@ -365,7 +460,7 @@ class FileHandler(object):
         if not self.confirm_destructive_action():
             return
         uri = action.get_current_uri()
-        fn = helpers.get_file_path_from_dnd_dropped_uri(uri)
+        fn = helpers.uri2filename(uri)
         self.open_file(fn)
 
     def open_last_cb(self, action):
@@ -375,13 +470,15 @@ class FileHandler(object):
         if not self.confirm_destructive_action():
             return
         uri = self.recent_items.pop().get_uri()
-        fn = helpers.get_file_path_from_dnd_dropped_uri(uri)
+        fn = helpers.uri2filename(uri)
         self.open_file(fn)
 
     def open_scrap_cb(self, action):
         groups = self.list_scraps_grouped()
         if not groups:
-            self.app.message_dialog(_('There are no scrap files named "%s" yet.' % (self.get_scrap_prefix() + '[0-9]*')), gtk.MESSAGE_WARNING)
+            msg = _('There are no scrap files named "%s" yet.') % \
+                (self.get_scrap_prefix() + '[0-9]*')
+            self.app.message_dialog(msg, gtk.MESSAGE_WARNING)
             return
         if not self.confirm_destructive_action():
             return
