@@ -29,16 +29,23 @@ class TiledDrawWidget(gtk.DrawingArea):
         self.connect("leave-notify-event", self.leave_notify_cb)
         self.connect("size-allocate", self.size_allocate_cb)
 
-        # workaround for https://gna.org/bugs/?14372
+        # workaround for https://gna.org/bugs/?14372 ([Windows] crash when moving the pen during startup)
         def at_application_start(*trash):
             self.connect("motion-notify-event", self.motion_notify_cb)
             self.connect("button-press-event", self.button_press_cb)
+            self.connect("button-release-event", self.button_release_cb)
         gobject.idle_add(at_application_start)
 
         self.set_events(gdk.EXPOSURE_MASK
                         | gdk.POINTER_MOTION_MASK
                         | gdk.ENTER_NOTIFY_MASK
                         | gdk.LEAVE_NOTIFY_MASK
+                        # Workaround for https://gna.org/bugs/index.php?16253
+                        # Mypaint doesn't use proximity-*-event for anything
+                        # yet, but this seems to be needed for scrollwheels
+                        # etc. to keep working.
+                        | gdk.PROXIMITY_OUT_MASK
+                        | gdk.PROXIMITY_IN_MASK
                         # for some reason we also need to specify events handled in drawwindow.py:
                         | gdk.BUTTON_PRESS_MASK
                         | gdk.BUTTON_RELEASE_MASK
@@ -56,6 +63,7 @@ class TiledDrawWidget(gtk.DrawingArea):
         self.last_event_x = None
         self.last_event_y = None
         self.last_event_device = None
+        self.last_event_had_pressure_info = False
         self.last_painting_pos = None
         self.device_observers = []
 
@@ -65,7 +73,7 @@ class TiledDrawWidget(gtk.DrawingArea):
         self.translation_y = 0.0
         self.scale = 1.0
         self.rotation = 0.0
-        self.flipped = False
+        self.mirrored = False
 
         self.has_pointer = False
         self.dragfunc = None
@@ -78,13 +86,21 @@ class TiledDrawWidget(gtk.DrawingArea):
         # gets overwritten for the main window
         self.zoom_max = 5.0
         self.zoom_min = 1/5.0
-        
+
         #self.scroll_at_edges = False
         self.pressure_mapping = None
+        self.bad_devices = []
+        self.motions = []
+
+        self.set_sensitive(True)
 
     #def set_scroll_at_edges(self, choice):
     #    self.scroll_at_edges = choice
-      
+
+    def set_sensitive(self, sensitive):
+        """Set if the widget accepts input or not"""
+        self.is_sensitive = sensitive
+
     def enter_notify_cb(self, widget, event):
         self.has_pointer = True
     def leave_notify_cb(self, widget, event):
@@ -100,11 +116,14 @@ class TiledDrawWidget(gtk.DrawingArea):
             dy = old_size[1] - new_size[1]
             self.scroll(dx/2, dy/2)
 
-    def motion_notify_cb(self, widget, event):
+    def motion_notify_cb(self, widget, event, button1_pressed=None):
+        if not self.is_sensitive:
+            return
+
         if self.last_event_time:
             dtime = (event.time - self.last_event_time)/1000.0
-            dx_int = int(event.x) - int(self.last_event_x)
-            dy_int = int(event.y) - int(self.last_event_y)
+            dx = event.x - self.last_event_x
+            dy = event.y - self.last_event_y
         else:
             dtime = None
         if event.device != self.last_event_device:
@@ -118,24 +137,44 @@ class TiledDrawWidget(gtk.DrawingArea):
             return
 
         if self.dragfunc:
-            if dx_int or dy_int:
-                # we only allow scrolling by full pixels, because it is much faster
-                self.dragfunc(dx_int, dy_int)
+            self.dragfunc(dx, dy)
             return
 
         cr = self.get_model_coordinates_cairo_context()
         x, y = cr.device_to_user(event.x, event.y)
         
         pressure = event.get_axis(gdk.AXIS_PRESSURE)
+
+        if pressure is not None and (pressure > 1.0 or pressure < 0.0):
+            if event.device.name not in self.bad_devices:
+                print 'WARNING: device "%s" is reporting bad pressure %+f' % (event.device.name, pressure)
+                self.bad_devices.append(event.device.name)
+            if pressure > 1000.0 or pressure < -1000.0:
+                # infinity: use button state (instead of clamping in brush.hpp)
+                # https://gna.org/bugs/?14709
+                pressure = None
+
         if pressure is None:
-            if event.state & gdk.BUTTON1_MASK:
+            self.last_event_had_pressure_info = False
+            if button1_pressed is None:
+                button1_pressed = event.state & gdk.BUTTON1_MASK
+            if button1_pressed:
                 pressure = 0.5
             else:
                 pressure = 0.0
+        else:
+            self.last_event_had_pressure_info = True
+
+        xtilt = event.get_axis(gdk.AXIS_XTILT)
+        ytilt = event.get_axis(gdk.AXIS_YTILT)
+        if xtilt is None or ytilt is None:
+            xtilt = 0.0
+            ytilt = 0.0
         
-        if event.state & gdk.CONTROL_MASK:
+        if event.state & gdk.CONTROL_MASK or event.state & gdk.MOD1_MASK:
             # color picking, do not paint
-            return
+            # Don't simply return; this is a workaround for unwanted lines in https://gna.org/bugs/?16169
+            pressure = 0.0
             
         ### CSS experimental - scroll when touching the edge of the screen in fullscreen mode
         #
@@ -165,12 +204,54 @@ class TiledDrawWidget(gtk.DrawingArea):
         if pressure:
             self.last_painting_pos = x, y
 
-        self.doc.stroke_to(dtime, x, y, pressure)
+        # On Windows, GTK timestamps have a resolution around
+        # 15ms, but tablet events arrive every 8ms.
+        # https://gna.org/bugs/index.php?16569
+        # TODO: proper fix in the brush engine, using only smooth,
+        #       filtered speed inputs, will make this unneccessary
+        if dtime < 0.0:
+            print 'Time is running backwards, dtime=%f' % dtime
+            dtime = 0.0
+        data = (x, y, pressure, xtilt, ytilt)
+        if dtime == 0.0:
+            self.motions.append(data)
+        elif dtime > 0.0:
+            if self.motions:
+                # replay previous events that had identical timestamp
+                if dtime > 0.1:
+                    # really old events, don't associate them with the new one
+                    step = 0.1
+                else:
+                    step = dtime
+                step /= len(self.motions)+1
+                for data_old in self.motions:
+                    self.doc.stroke_to(step, *data_old)
+                    dtime -= step
+                self.motions = []
+            self.doc.stroke_to(dtime, *data)
 
     def button_press_cb(self, win, event):
-        if (event.state & gdk.SHIFT_MASK) and self.last_painting_pos:
-            dst = self.get_cursor_in_model_coordinates()
-            self.doc.straight_line(self.last_painting_pos, dst)
+        if event.type != gdk.BUTTON_PRESS:
+            # ignore the extra double-click event
+            return
+
+        if event.button == 1:
+            # straight line
+            if (event.state & gdk.SHIFT_MASK) and self.last_painting_pos:
+                dst = self.get_cursor_in_model_coordinates()
+                self.doc.straight_line(self.last_painting_pos, dst)
+
+            # mouse button pressed (while painting without pressure information)
+            if not self.last_event_had_pressure_info:
+                # For the mouse we don't get a motion event for "pressure"
+                # changes, so we simulate it. (Note: we can't use the
+                # event's button state because it carries the old state.)
+                self.motion_notify_cb(win, event, button1_pressed=True)
+
+    def button_release_cb(self, win, event):
+        # (see comment above in button_press_cb)
+        if event.button == 1 and not self.last_event_had_pressure_info:
+            self.motion_notify_cb(win, event, button1_pressed=False)
 
     def canvas_modified_cb(self, x1, y1, w, h):
         if not self.window:
@@ -202,25 +283,38 @@ class TiledDrawWidget(gtk.DrawingArea):
         return True
 
     def get_model_coordinates_cairo_context(self, cr=None):
+        # OPTIMIZE: check whether this is a bottleneck during painting (many motion events) - if yes, use cache
         if cr is None:
             cr = self.window.cairo_create()
+
+        scale = self.scale
+        # check if scale is almost a power of two
+        scale_log2 = log(scale, 2)
+        scale_log2_rounded = round(scale_log2)
+        if abs(scale_log2-scale_log2_rounded) < 0.01:
+            scale = 2.0**scale_log2_rounded
+
+        rotation = self.rotation # maybe we should check if rotation is almost a multiple of 90 degrees?
+
         cr.translate(self.translation_x, self.translation_y)
-        cr.rotate(self.rotation)
-        cr.scale(self.scale, self.scale)
-        if self.flipped:
+        cr.rotate(rotation)
+        cr.scale(scale, scale)
+
+        # Align the translation such that (0,0) maps to an integer
+        # screen pixel, to keep image rendering fast and sharp.
+        x, y = cr.user_to_device(0, 0)
+        x, y = cr.device_to_user(round(x), round(y))
+        cr.translate(x, y)
+
+        if self.mirrored:
             m = list(cr.get_matrix())
             m[0] = -m[0]
             m[2] = -m[2]
             cr.set_matrix(cairo.Matrix(*m))
-        # does not seem to make a difference:
-        #cr.set_antialias(cairo.ANTIALIAS_SUBPIXEL)
-        # this one neither:
-        #cr.set_antialias(cairo.ANTIALIAS_NONE)
-        # looks like we always get nearest-neighbour downsampling
         return cr
 
     def is_translation_only(self):
-        return self.rotation == 0.0 and self.scale == 1.0 and not self.flipped
+        return self.rotation == 0.0 and self.scale == 1.0 and not self.mirrored
 
     def get_cursor_in_model_coordinates(self):
         x, y, modifiers = self.window.get_pointer()
@@ -263,6 +357,8 @@ class TiledDrawWidget(gtk.DrawingArea):
 
         # choose best mipmap
         mipmap_level = max(0, int(ceil(log(1/self.scale,2))))
+        #mipmap_level = max(0, int(floor(log(1.0/self.scale,2)))) # slightly better quality but clearly slower
+        # OPTIMIZE: if we would render tile scanlines, we could probably use the better one above...
         mipmap_level = min(mipmap_level, tiledsurface.MAX_MIPMAP_LEVEL)
         cr.scale(2**mipmap_level, 2**mipmap_level)
 
@@ -280,7 +376,11 @@ class TiledDrawWidget(gtk.DrawingArea):
         x1, y1 = int(floor(x1)), int(floor(y1))
         x2, y2 = int(ceil (x2)), int(ceil (y2))
 
-        surface = pixbufsurface.Surface(x1, y1, x2-x1+1, y2-y1+1)
+        # alpha=True is just to get hardware acceleration, we don't
+        # actually use the alpha channel. Speedup factor 3 for
+        # ATI/Radeon Xorg driver (and hopefully others).
+        # https://bugs.freedesktop.org/show_bug.cgi?id=28670
+        surface = pixbufsurface.Surface(x1, y1, x2-x1+1, y2-y1+1, alpha=True)
 
         del x1, y1, x2, y2, w, h
 
@@ -337,7 +437,21 @@ class TiledDrawWidget(gtk.DrawingArea):
             x, y = cr.user_to_device(surface.x, surface.y)
             self.window.draw_pixbuf(None, surface.pixbuf, 0, 0, int(x), int(y), dither=gdk.RGB_DITHER_MAX)
         else:
-            cr.set_source_pixbuf(surface.pixbuf, surface.x, surface.y)
+            #print 'Position (screen coordinates):', cr.user_to_device(surface.x, surface.y)
+            cr.set_source_pixbuf(surface.pixbuf, round(surface.x), round(surface.y))
+            pattern = cr.get_source()
+
+            # We could set interpolation mode here (eg nearest neighbour)
+            #pattern.set_filter(cairo.FILTER_NEAREST)  # 1.6s
+            #pattern.set_filter(cairo.FILTER_FAST)     # 2.0s
+            #pattern.set_filter(cairo.FILTER_GOOD)     # 3.1s
+            #pattern.set_filter(cairo.FILTER_BEST)     # 3.1s
+            #pattern.set_filter(cairo.FILTER_BILINEAR) # 3.1s
+
+            if self.scale > 3.0:
+                # pixelize at high zoom-in levels
+                pattern.set_filter(cairo.FILTER_NEAREST)
+
             cr.paint()
 
         if self.visualize_rendering:
@@ -345,11 +459,9 @@ class TiledDrawWidget(gtk.DrawingArea):
             cr.set_source_rgba(0, 0, random.random(), 0.4)
             cr.paint()
 
-    def scroll(self, dx, dy, show_immediately=True):
-        assert int(dx) == dx and int(dy) == dy
+    def scroll(self, dx, dy):
         self.translation_x -= dx
         self.translation_y -= dy
-        #if show_immediately:
         if False:
             # This speeds things up nicely when scrolling is already
             # fast, but produces temporary artefacts and an
@@ -363,7 +475,7 @@ class TiledDrawWidget(gtk.DrawingArea):
         else:
             self.queue_draw()
 
-    def rotozoom_with_center(self, function, at_pointer=True):
+    def rotozoom_with_center(self, function, at_pointer=False):
         if at_pointer and self.has_pointer and self.last_event_x is not None:
             cx, cy = self.last_event_x, self.last_event_y
         else:
@@ -378,21 +490,15 @@ class TiledDrawWidget(gtk.DrawingArea):
         self.translation_x += cx - cx_new
         self.translation_y += cy - cy_new
 
-        # this is for fast scrolling with only tanslation
-        self.rotation = self.rotation % (2*pi)
-        if self.is_translation_only():
-            self.translation_x = int(self.translation_x)
-            self.translation_y = int(self.translation_y)
-
         self.queue_draw()
 
     def zoom(self, zoom_step):
         def f(): self.scale *= zoom_step
-        self.rotozoom_with_center(f)
+        self.rotozoom_with_center(f, at_pointer=True)
 
     def set_zoom(self, zoom):
         def f(): self.scale = zoom
-        self.rotozoom_with_center(f)
+        self.rotozoom_with_center(f, at_pointer=True)
 
     def rotate(self, angle_step):
         def f(): self.rotation += angle_step
@@ -402,9 +508,13 @@ class TiledDrawWidget(gtk.DrawingArea):
         def f(): self.rotation = angle
         self.rotozoom_with_center(f)
 
-    def set_flipped(self, flipped):
-        def f(): self.flipped = flipped
-        self.rotozoom_with_center(f, at_pointer=False)
+    def mirror(self):
+        def f(): self.mirrored = not self.mirrored
+        self.rotozoom_with_center(f)
+
+    def set_mirrored(self, mirrored):
+        def f(): self.mirrored = mirrored
+        self.rotozoom_with_center(f)
 
     def start_drag(self, dragfunc):
         self.dragfunc = dragfunc
@@ -421,9 +531,8 @@ class TiledDrawWidget(gtk.DrawingArea):
         w, h = self.window.get_size()
         cx_user, cy_user = cr.device_to_user(w/2, h/2)
 
-        # note: integer rounding above avoids fractional translation
-        self.translation_x += cx_user - desired_cx_user
-        self.translation_y += cy_user - desired_cy_user
+        self.translation_x += (cx_user - desired_cx_user)*self.scale
+        self.translation_y += (cy_user - desired_cy_user)*self.scale
         self.queue_draw()
 
     def brush_modified_cb(self):
