@@ -8,6 +8,7 @@
 
 import os, zipfile, tempfile, time, traceback
 join = os.path.join
+from cStringIO import StringIO
 import xml.etree.ElementTree as ET
 from gtk import gdk
 import gobject, numpy
@@ -97,16 +98,16 @@ class Document():
         if not self.layer.surface.is_empty():
             self.do(command.ClearLayer(self))
 
-    def stroke_to(self, dtime, x, y, pressure):
+    def stroke_to(self, dtime, x, y, pressure, xtilt,ytilt):
         if not self.stroke:
             self.stroke = stroke.Stroke()
             self.stroke.start_recording(self.brush)
             self.snapshot_before_stroke = self.layer.save_snapshot()
-        self.stroke.record_event(dtime, x, y, pressure)
+        self.stroke.record_event(dtime, x, y, pressure, xtilt,ytilt)
 
         l = self.layer
         l.surface.begin_atomic()
-        split = self.brush.stroke_to (l.surface, x, y, pressure, dtime)
+        split = self.brush.stroke_to (l.surface, x, y, pressure, xtilt,ytilt, dtime)
         l.surface.end_atomic()
 
         if split:
@@ -127,9 +128,9 @@ class Document():
         # rest the brush in src for a minute, to avoid interpolation
         # from the upper left corner (states are zero) (FIXME: the
         # brush should handle this on its own, maybe?)
-        self.stroke_to(60.0, x[0], y[0], 0.0)
+        self.stroke_to(60.0, x[0], y[0], 0.0, 0.0, 0.0)
         for i in xrange(N):
-            self.stroke_to(duration/N, x[i], y[i], pressure)
+            self.stroke_to(duration/N, x[i], y[i], pressure, 0.0, 0.0)
         self.split_stroke()
         self.brush = real_brush
 
@@ -178,7 +179,7 @@ class Document():
             res.expandToIncludeRect(bbox)
         return res
 
-    def blit_tile_into(self, dst_8bit, tx, ty, mipmap=0, layers=None, background=None):
+    def blit_tile_into(self, dst_8bit, tx, ty, mipmap_level=0, layers=None, background=None):
         if layers is None:
             layers = self.layers
         if background is None:
@@ -187,11 +188,11 @@ class Document():
         assert dst_8bit.dtype == 'uint8'
         dst = numpy.empty((N, N, 3), dtype='uint16')
 
-        background.blit_tile_into(dst, tx, ty, mipmap)
+        background.blit_tile_into(dst, tx, ty, mipmap_level)
 
         for layer in layers:
             surface = layer.surface
-            surface.composite_tile_over(dst, tx, ty, mipmap_level=mipmap, opacity=layer.effective_opacity)
+            surface.composite_tile_over(dst, tx, ty, mipmap_level=mipmap_level, opacity=layer.effective_opacity)
 
         mypaintlib.tile_convert_rgb16_to_rgb8(dst, dst_8bit)
 
@@ -215,11 +216,18 @@ class Document():
         arr = helpers.gdkpixbuf2numpy(pixbuf)
         self.do(command.LoadLayer(self, arr, x, y))
 
-    def set_layer_opacity(self, opacity):
+    def set_layer_visibility(self, visible, layer):
+        cmd = self.get_last_command()
+        if isinstance(cmd, command.SetLayerVisibility) and cmd.layer is layer:
+            self.undo()
+        self.do(command.SetLayerVisibility(self, visible, layer))
+
+    def set_layer_opacity(self, opacity, layer=None):
+        """Sets the opacity of a layer. If layer=None, works on the current layer"""
         cmd = self.get_last_command()
         if isinstance(cmd, command.SetLayerOpacity):
             self.undo()
-        self.do(command.SetLayerOpacity(self, opacity))
+        self.do(command.SetLayerOpacity(self, opacity, layer))
 
     def set_background(self, obj):
         # This is not an undoable action. One reason is that dragging
@@ -287,23 +295,22 @@ class Document():
     def unsupported(self, filename):
         raise SaveLoadError, _('Unknown file format extension: %s') % repr(filename)
 
-    def render_as_pixbuf(self, *args):
-        return pixbufsurface.render_as_pixbuf(self, *args)
+    def render_as_pixbuf(self, *args, **kwargs):
+        return pixbufsurface.render_as_pixbuf(self, *args, **kwargs)
 
-    def save_png(self, filename, compression=2, alpha=False, multifile=False):
+    def save_png(self, filename, alpha=False, multifile=False):
         if multifile:
-            self.save_multifile_png(filename, compression)
+            self.save_multifile_png(filename)
         else:
             if alpha:
                 tmp_layer = layer.Layer()
                 for l in self.layers:
                     l.merge_into(tmp_layer)
-                pixbuf = tmp_layer.surface.render_as_pixbuf()
+                tmp_layer.surface.save(filename)
             else:
-                pixbuf = self.render_as_pixbuf()
-            pixbuf.save(filename, 'png', {'compression':str(compression)})
+                pixbufsurface.save_as_png(self, filename, alpha=False)
 
-    def save_multifile_png(self, filename, compression=2, alpha=False):
+    def save_multifile_png(self, filename, alpha=False):
         prefix, ext = os.path.splitext(filename)
         # if we have a number already, strip it
         l = prefix.rsplit('.', 1)
@@ -348,15 +355,23 @@ class Document():
         def store_pixbuf(pixbuf, name):
             tmp = join(tempdir, 'tmp.png')
             t1 = time.time()
-            pixbuf.save(tmp, 'png', {'compression':'2'})
-            print '  %.3fs saving %s compression 2' % (time.time() - t1, name)
+            pixbuf.save(tmp, 'png')
+            print '  %.3fs pixbuf saving %s' % (time.time() - t1, name)
             z.write(tmp, name)
             os.remove(tmp)
 
-        def add_layer(x, y, opac, pixbuf, name, layer_name):
+        def store_surface(surface, name, rect=[]):
+            tmp = join(tempdir, 'tmp.png')
+            t1 = time.time()
+            surface.save(tmp, *rect)
+            print '  %.3fs surface saving %s' % (time.time() - t1, name)
+            z.write(tmp, name)
+            os.remove(tmp)
+
+        def add_layer(x, y, opac, surface, name, layer_name, visible=True, rect=[]):
             layer = ET.Element('layer')
             stack.append(layer)
-            store_pixbuf(pixbuf, name)
+            store_surface(surface, name, rect)
             a = layer.attrib
             if layer_name:
                 a['name'] = layer_name
@@ -364,6 +379,10 @@ class Document():
             a['x'] = str(x)
             a['y'] = str(y)
             a['opacity'] = str(opac)
+            if visible:
+                a['visibility'] = 'visible'
+            else:
+                a['visibility'] = 'hidden'
             return layer
 
         for idx, l in enumerate(reversed(self.layers)):
@@ -371,29 +390,36 @@ class Document():
                 continue
             opac = l.opacity
             x, y, w, h = l.surface.get_bbox()
-            pixbuf = l.surface.render_as_pixbuf()
-            el = add_layer(x-x0, y-y0, opac, pixbuf, 'data/layer%03d.png' % idx, l.name)
+            el = add_layer(x-x0, y-y0, opac, l.surface, 'data/layer%03d.png' % idx, l.name, l.visible, rect=(x, y, w, h))
             # strokemap
-            data = l.save_strokemap_to_string(-x, -y)
+            sio = StringIO()
+            l.save_strokemap_to_file(sio, -x, -y)
+            data = sio.getvalue(); sio.close()
             name = 'data/layer%03d_strokemap.dat' % idx
-            el.attrib['mypaint_strokemap'] = name
+            el.attrib['mypaint_strokemap_v2'] = name
             write_file_str(name, data)
 
         # save background as layer (solid color or tiled)
-        s = pixbufsurface.Surface(x0, y0, w0, h0)
-        s.fill(self.background)
-        l = add_layer(0, 0, 1.0, s.pixbuf, 'data/background.png', 'background')
         bg = self.background
+        # save as fully rendered layer
+        l = add_layer(0, 0, 1.0, bg, 'data/background.png', 'background', rect=(x0, y0, w0, h0))
         x, y, w, h = bg.get_pattern_bbox()
-        pixbuf = pixbufsurface.render_as_pixbuf(bg, x+x0, y+y0, w, h, alpha=False)
-        store_pixbuf(pixbuf, 'data/background_tile.png')
+        # save as single pattern (with corrected origin)
+        store_surface(bg, 'data/background_tile.png', rect=(x+x0, y+y0, w, h))
         l.attrib['background_tile'] = 'data/background_tile.png'
 
-        # preview
+        # preview (256x256)
         t2 = time.time()
-        print '  starting to render image for thumbnail...'
-        pixbuf = self.render_as_pixbuf()
-        w, h = pixbuf.get_width(), pixbuf.get_height()
+        print '  starting to render full image for thumbnail...'
+
+        x, y, w, h = x0, y0, w0, h0
+        mipmap_level = 0
+        while mipmap_level < tiledsurface.MAX_MIPMAP_LEVEL and max(w, h) >= 512:
+            mipmap_level += 1
+            x, y, w, h = x/2, y/2, w/2, h/2
+
+        pixbuf = self.render_as_pixbuf(x, y, w, h, mipmap_level=mipmap_level)
+        assert pixbuf.get_width() == w and pixbuf.get_height() == h
         if w > h:
             w, h = 256, max(h*256/w, 1)
         else:
@@ -417,6 +443,7 @@ class Document():
         print '%.3fs save_ora total' % (time.time() - t0)
 
     def load_ora(self, filename):
+        """Loads from an OpenRaster file"""
         print 'load_ora:'
         t0 = time.time()
         tempdir = tempfile.mkdtemp('mypaint')
@@ -478,23 +505,25 @@ class Document():
             x = int(a.get('x', '0'))
             y = int(a.get('y', '0'))
             opac = float(a.get('opacity', '1.0'))
-
+            visible = not 'hidden' in a.get('visibility', 'visible')
             self.add_layer(insert_idx=0, name=name)
             last_pixbuf = pixbuf
             t1 = time.time()
             self.load_layer_from_pixbuf(pixbuf, x, y)
             layer = self.layers[0]
 
-            layer.opacity = helpers.clamp(opac, 0.0, 1.0)
+            self.set_layer_opacity(helpers.clamp(opac, 0.0, 1.0), layer)
+            self.set_layer_visibility(visible, layer)
             print '  %.3fs converting pixbuf to layer format' % (time.time() - t1)
             # strokemap
-            fname = a.get('mypaint_strokemap', None)
+            fname = a.get('mypaint_strokemap_v2', None)
             if fname:
                 if x % N or y % N:
                     print 'Warning: dropping non-aligned strokemap'
                 else:
-                    data = z.read(fname)
-                    self.layers[0].load_strokemap_from_string(data, x, y)
+                    sio = StringIO(z.read(fname))
+                    layer.load_strokemap_from_file(sio, x, y)
+                    sio.close()
 
         os.rmdir(tempdir)
 
